/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include "vec4.hpp"

#include "vec3.hpp"
#include "vec2.hpp"
#include "../libcommon/exception_cpe.hpp"

#include <cmath>
#include <sstream>

#include <assert.h>

namespace cpe
{

vec4::vec4():internal_x(0),internal_y(0),internal_z(0),internal_w(0){}
vec4::vec4(float x_param,float y_param,float z_param,float w_param):internal_x(x_param),internal_y(y_param),internal_z(z_param),internal_w(w_param){}
float vec4::x() const
{
    return internal_x;
}

float& vec4::x()
{
    return internal_x;
}

float vec4::y() const
{
    return internal_y;
}

float& vec4::y()
{
    return internal_y;
}

float vec4::z() const
{
    return internal_z;
}

float& vec4::z()
{
    return internal_z;
}
float vec4::w() const
{
    return internal_w;
}

float& vec4::w()
{
    return internal_w;
}

float vec4::operator[](const size_t& k) const
{
    assert_size(k);
    switch (k)
    {
    case 0:
        return internal_x;
    case 1:
        return internal_y;
    case 2:
        return internal_z;
    case 3:
        return internal_w;
    default:
        assert(false);
        throw cpe::exception_cpe("Probleme",MACRO_EXCEPTION_PARAMETER);
    };
}

float& vec4::operator[](const size_t& k)
{
    assert_size(k);
    switch (k)
    {
    case 0:
        return internal_x;
    case 1:
        return internal_y;
    case 2:
        return internal_z;
    case 3:
        return internal_w;
    default:
        assert(false);
        throw cpe::exception_cpe("Probleme",MACRO_EXCEPTION_PARAMETER);
    };
}

float vec4::operator()(const size_t& k) const
{
    return (*this)[k];
}

float& vec4::operator()(const size_t& k)
{
    return (*this)[k];
}

void vec4::set_zero()
{
    internal_x=0;
    internal_y=0;
    internal_z=0;
    internal_w=0;
}

float vec4::dot(const vec4& p) const
{
    return internal_x*p.x()+internal_y*p.y()+internal_z*p.z()+internal_w*p.w();
}

float vec4::norm() const
{
    return std::sqrt(norm2());
}

float vec4::norm2() const
{
    return internal_x*internal_x+internal_y*internal_y+internal_z*internal_z+internal_w*internal_w;
}


vec4 vec4::operator+(const vec4& p2) const
{
    return vec4(x()+p2.x(),y()+p2.y(),z()+p2.z(),w()+p2.w());
}

vec4 operator+(float s,const vec4& p)
{
    return vec4(s+p.x(),s+p.y(),s+p.z(),s+p.w());
}

vec4 vec4::operator+(float s) const
{
    vec4 temp=*this;
    temp+=s;
    return temp;
}

vec4 vec4::operator-(const vec4& p2) const
{
    return vec4(x()-p2.x(),y()-p2.y(),z()-p2.z(),w()-p2.w());
}

vec4 operator-(float s,const vec4& p)
{
    return vec4(s-p.x(),s-p.y(),s-p.z(),s-p.w());
}

vec4 vec4::operator-(float s) const
{
    return vec4(x()-s,y()-s,z()-s,w()-s);
}

vec4 operator*(float s,const vec4& p)
{
    return vec4(p.x()*s,p.y()*s,p.z()*s,p.w()*s);
}

vec4 vec4::operator*(float s) const
{
    vec4 temp=*this;
    temp*=s;
    return temp;
}

vec4 vec4::operator/(float s) const
{
    float epsilon=1e-8;
    if(std::abs(s)<epsilon)
        throw cpe::exception_cpe("Division par 0",MACRO_EXCEPTION_PARAMETER);
    return vec4(x()/s,y()/s,z()/s,w()/s);
}

vec4& vec4::operator+=(const vec4& p)
{
    internal_x+=p.x();
    internal_y+=p.y();
    internal_z+=p.z();
    internal_w+=p.w();
    return *this;
}

vec4& vec4::operator+=(float s)
{
    internal_x+=s;
    internal_y+=s;
    internal_z+=s;
    internal_w+=s;
    return *this;
}

vec4& vec4::operator-=(const vec4& p)
{
    internal_x-=p.x();
    internal_y-=p.y();
    internal_z-=p.z();
    internal_w-=p.w();

    return *this;
}

vec4& vec4::operator-=(float s)
{
    internal_x-=s;
    internal_y-=s;
    internal_z-=s;
    internal_w-=s;
    return *this;
}

vec4& vec4::operator*=(float s)
{
    internal_x*=s;
    internal_y*=s;
    internal_z*=s;
    internal_w*=s;
    return *this;
}

vec4& vec4::operator/=(float s)
{
    float epsilon=1e-8;
    if(std::abs(s)<epsilon)
        throw cpe::exception_cpe("Division par 0",MACRO_EXCEPTION_PARAMETER);

    internal_x/=s;
    internal_y/=s;
    internal_z/=s;
    internal_w/=s;
    return *this;
}

vec4 vec4::operator-() const
{
    return vec4(-internal_x,-internal_y,-internal_z,-internal_w);
}

vec4 vec4::product_compontentwise(const vec4& p) const
{
    return vec4(internal_x*p.x(),internal_y*p.y(),internal_z*p.z(),internal_w*p.w());
}

vec4& vec4::product_compontentwise_internal(const vec4& p)
{
    internal_x*=p.x();
    internal_y*=p.y();
    internal_z*=p.z();
    internal_w*=p.w();
    return *this;
}
std::string vec4::to_string() const
{
    std::stringstream stream;
    stream<<internal_x<<" "<<internal_y<<" "<<internal_z<<" "<<internal_w;
    return stream.str();
}
std::ostream& operator<<(std::ostream& stream,const vec4& p)
{
    stream<<"("<<p.to_string()<<")";
    return stream;
}

void vec4::assert_size(const size_t& k) const
{
    if(k>3)
    {
        std::stringstream stream;
        stream<<"vec4::assert_size("<<k<<") k>3";
        throw exception_cpe(stream.str(),MACRO_EXCEPTION_PARAMETER);
    }
}
vec4 vec4::normalized() const
{
    float epsilon=1e-6;

    float n=norm();
    if(n<epsilon)
        throw exception_cpe("norm=0",MACRO_EXCEPTION_PARAMETER);
    return *this/n;
}
void vec4::scale(float sx,float sy,float sz,float sw)
{
    internal_x*=sx;
    internal_y*=sy;
    internal_z*=sz;
    internal_w*=sw;
}

vec3 vec4::to_vec3() const{return vec3(internal_x,internal_y,internal_z);}
vec2 vec4::to_vec2() const{return vec2(internal_x,internal_y);}

const float *vec4::pointer() const
{return &internal_x;}

}
